@@ -2,7 +2,7 @@
  *
  * common.h
  *
- * Copyright (c) 2010-2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2010-2012, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  *
  *-------------------------------------------------------------------------
  */
@@ -32,6 +32,14 @@
 #define LOGMSG_MAINTENANCE	LOG_PREFIX "maintenance requested"
 #define LOGMSG_RESTART		LOG_PREFIX "restart requested"
 
+/* maintenance mode flag */
+#define MAINTENANCE_MODE_SNAPSHOT	0x01
+#define MAINTENANCE_MODE_LOG		0x02
+
+/* exit code for pg_statsinfod */
+#define STATSINFO_EXIT_SUCCESS		0x00
+#define STATSINFO_EXIT_FAILED		0xff
+
 extern bool readControlFile(ControlFileData *ctrl, const char *pgdata);
 
 #endif   /* PG_STATSINFO_COMMON_H */
