@@ -1,7 +1,7 @@
 /*
  * report.c
  *
- * Copyright (c) 2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2011-2012, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 
 #include "pg_statsinfo.h"
@@ -20,8 +20,12 @@
 #define SQL_SELECT_RECOVERY_CONFLICTS 			"SELECT * FROM statsrepo.get_recovery_conflicts($1, $2)"
 #define SQL_SELECT_INSTANCE_PROC_RATIO			"SELECT * FROM statsrepo.get_proc_ratio($1, $2)"
 #define SQL_SELECT_INSTANCE_PROC_TENDENCY		"SELECT * FROM statsrepo.get_proc_tendency_report($1, $2)"
+#define SQL_SELECT_WALSTATS						"SELECT * FROM statsrepo.get_xlog_stats($1, $2)"
+#define SQL_SELECT_WALSTATS_TENDENCY			"SELECT * FROM statsrepo.get_xlog_tendency($1, $2)"
 #define SQL_SELECT_CPU_USAGE					"SELECT * FROM statsrepo.get_cpu_usage($1, $2)"
 #define SQL_SELECT_CPU_USAGE_TENDENCY			"SELECT * FROM statsrepo.get_cpu_usage_tendency_report($1, $2)"
+#define SQL_SELECT_LOADAVG_TENDENCY				"SELECT * FROM statsrepo.get_loadavg_tendency($1, $2)"
+#define SQL_SELECT_MEMORY_TENDENCY				"SELECT * FROM statsrepo.get_memory_tendency($1, $2)"
 #define SQL_SELECT_IO_USAGE						"SELECT * FROM statsrepo.get_io_usage($1, $2)"
 #define SQL_SELECT_IO_USAGE_TENDENCY			"SELECT * FROM statsrepo.get_io_usage_tendency_report($1, $2)"
 #define SQL_SELECT_DISK_USAGE_TABLESPACE		"SELECT * FROM statsrepo.get_disk_usage_tablespace($1, $2)"
@@ -32,10 +36,30 @@
 #define SQL_SELECT_LOW_DENSITY_TABLES			"SELECT * FROM statsrepo.get_low_density_tables($1, $2) LIMIT 10"
 #define SQL_SELECT_FRAGMENTED_TABLES			"SELECT * FROM statsrepo.get_flagmented_tables($1, $2)"
 #define SQL_SELECT_CHECKPOINT_ACTIVITY			"SELECT * FROM statsrepo.get_checkpoint_activity($1, $2)"
-#define SQL_SELECT_AUTOVACUUM_ACTIVITY			"SELECT * FROM statsrepo.get_autovacuum_activity($1, $2)"
+#define SQL_SELECT_AUTOVACUUM_ACTIVITY "\
+SELECT \
+	datname || '.' || nspname || '.' || relname, \
+	\"count\", \
+	avg_index_scans, \
+	avg_tup_removed, \
+	avg_tup_remain, \
+	avg_duration, \
+	max_duration \
+FROM \
+	statsrepo.get_autovacuum_activity($1, $2)"
+#define SQL_SELECT_AUTOVACUUM_ACTIVITY2 "\
+SELECT \
+	datname || '.' || nspname || '.' || relname, \
+	avg_page_hit, \
+	avg_page_miss, \
+	avg_page_dirty, \
+	avg_read_rate, \
+	avg_write_rate \
+FROM \
+	statsrepo.get_autovacuum_activity2($1, $2)"
 #define SQL_SELECT_QUERY_ACTIVITY_FUNCTIONS		"SELECT * FROM statsrepo.get_query_activity_functions($1, $2) LIMIT 20"
 #define SQL_SELECT_QUERY_ACTIVITY_STATEMENTS	"SELECT * FROM statsrepo.get_query_activity_statements($1, $2) LIMIT 20"
-#define SQL_SELECT_LOCK_ACTIVITY				"SELECT * FROM statsrepo.get_lock_activity($1, $2) LIMIT 20"
+#define SQL_SELECT_LOCK_CONFLICTS				"SELECT * FROM statsrepo.get_lock_activity($1, $2) LIMIT 20"
 #define SQL_SELECT_REPLICATION_ACTIVITY			"SELECT * FROM statsrepo.get_replication_activity($1, $2)"
 #define SQL_SELECT_SETTING_PARAMETERS			"SELECT * FROM statsrepo.get_setting_parameters($1, $2)"
 #define SQL_SELECT_SCHEMA_INFORMATION_TABLES	"SELECT * FROM statsrepo.get_schema_info_tables($1, $2)"
@@ -47,6 +71,7 @@
 		i.instid, \
 		i.hostname, \
 		i.port, \
+		i.pg_version, \
 		min(s.snapid), \
 		max(s.snapid) \
 	FROM \
@@ -57,7 +82,8 @@
 	GROUP BY \
 		i.instid, \
 		i.hostname, \
-		i.port \
+		i.port, \
+		i.pg_version \
 	ORDER BY \
 		i.instid"
 
@@ -66,6 +92,7 @@
 		i.instid, \
 		i.hostname, \
 		i.port, \
+		i.pg_version, \
 		min(s.snapid), \
 		max(s.snapid) \
 	FROM \
@@ -76,18 +103,20 @@
 	GROUP BY \
 		i.instid, \
 		i.hostname, \
-		i.port \
+		i.port, \
+		i.pg_version \
 	ORDER BY \
 		i.instid"
 
 /* the report scope per instance */
 typedef struct ReportScope
 {
-	char	*instid;	/* instance ID */
-	char	*host;		/* host */
-	char	*port;		/* port */
-	char	*beginid;	/* begin point of report */
-	char	*endid;		/* end point of report */
+	char	*instid;		/* instance ID */
+	char	*host;			/* host */
+	char	*port;			/* port */
+	int		 version;		/* PostgreSQL version */
+	char	*beginid;		/* begin point of report */
+	char	*endid;			/* end point of report */
 } ReportScope;
 
 /* function interface of the report builder */
@@ -104,7 +133,7 @@ static void report_notable_tables(PGconn *conn, ReportScope *scope, FILE *out);
 static void report_checkpoint_activity(PGconn *conn, ReportScope *scope, FILE *out);
 static void report_autovacuum_activity(PGconn *conn, ReportScope *scope, FILE *out);
 static void report_query_activity(PGconn *conn, ReportScope *scope, FILE *out);
-static void report_lock_activity(PGconn *conn, ReportScope *scope, FILE *out);
+static void report_lock_conflicts(PGconn *conn, ReportScope *scope, FILE *out);
 static void report_replication_activity(PGconn *conn, ReportScope *scope, FILE *out);
 static void report_setting_parameters(PGconn *conn, ReportScope *scope, FILE *out);
 static void report_schema_information(PGconn *conn, ReportScope *scope, FILE *out);
@@ -116,6 +145,7 @@ static List *select_scope_by_snapid(PGconn *conn, const char *beginid, const cha
 static List *select_scope_by_timestamp(PGconn *conn, time_t begindate, time_t enddate);
 static void destroy_report_scope(ReportScope *scope);
 static int get_server_version(PGconn *conn);
+static int parse_version(const char *versionString);
 
 /*
  * generate a report
@@ -266,14 +296,30 @@ report_database_statistics(PGconn *conn, ReportScope *scope, FILE *out)
 	for (i = 0; i < PQntuples(res); i++)
 	{
 		fprintf(out, "Database Name              : %s\n", PQgetvalue(res, i, 0));
-		fprintf(out, "Database Size              : %s MB\n", PQgetvalue(res, i, 1));
-		fprintf(out, "Database Size Increase     : %s MB\n", PQgetvalue(res, i, 2));
+		fprintf(out, "Database Size              : %s MiB\n", PQgetvalue(res, i, 1));
+		fprintf(out, "Database Size Increase     : %s MiB\n", PQgetvalue(res, i, 2));
 		fprintf(out, "Commit/s                   : %s\n", PQgetvalue(res, i, 3));
 		fprintf(out, "Rollback/s                 : %s\n", PQgetvalue(res, i, 4));
 		fprintf(out, "Cache Hit Ratio            : %s %%\n", PQgetvalue(res, i, 5));
 		fprintf(out, "Block Read/s (disk+cache)  : %s\n", PQgetvalue(res, i, 6));
 		fprintf(out, "Block Read/s (disk)        : %s\n", PQgetvalue(res, i, 7));
-		fprintf(out, "Rows Read/s                : %s\n\n", PQgetvalue(res, i, 8));
+		fprintf(out, "Rows Read/s                : %s\n", PQgetvalue(res, i, 8));
+		if (scope->version >= 90200)
+		{
+			fprintf(out, "Temporary Files            : %s\n", PQgetvalue(res, i, 9));
+			fprintf(out, "Temporary Bytes            : %s MiB\n", PQgetvalue(res, i, 10));
+			fprintf(out, "Deadlocks                  : %s\n", PQgetvalue(res, i, 11));
+			fprintf(out, "Block Read Time            : %s ms\n", PQgetvalue(res, i, 12));
+			fprintf(out, "Block Write Time           : %s ms\n\n", PQgetvalue(res, i, 13));
+		}
+		else
+		{
+			fprintf(out, "Temporary Files            : (N/A)\n");
+			fprintf(out, "Temporary Bytes            : (N/A)\n");
+			fprintf(out, "Deadlocks                  : (N/A)\n");
+			fprintf(out, "Block Read Time            : (N/A)\n");
+			fprintf(out, "Block Write Time           : (N/A)\n\n");
+		}
 	}
 	PQclear(res);
 
@@ -297,14 +343,14 @@ report_database_statistics(PGconn *conn, ReportScope *scope, FILE *out)
 
 	fprintf(out, "/** Database Size **/\n");
 	fprintf(out, "-----------------------------------\n");
-	fprintf(out, "%-16s  %-16s  %13s\n",
+	fprintf(out, "%-16s  %-16s  %14s\n",
 		"DateTime", "Database", "Size");
-	fprintf(out, "----------------------------------------------------\n");
+	fprintf(out, "-----------------------------------------------------\n");
 
 	res = pgut_execute(conn, SQL_SELECT_DBSIZE_TENDENCY, lengthof(params), params);
 	for(i = 0; i < PQntuples(res); i++)
 	{
-		fprintf(out, "%-16s  %-16s  %10s MB\n",
+		fprintf(out, "%-16s  %-16s  %10s MiB\n",
 			PQgetvalue(res, i, 0),
 			PQgetvalue(res, i, 1),
 			PQgetvalue(res, i, 2));
@@ -319,19 +365,22 @@ report_database_statistics(PGconn *conn, ReportScope *scope, FILE *out)
 		"Conflict Snapshot", "Conflict Bufferpin", "Conflict Deadlock");
 	fprintf(out, "-----------------------------------------------------------------------------------------------------------------\n");
 
-	res = pgut_execute(conn, SQL_SELECT_RECOVERY_CONFLICTS, lengthof(params), params);
-	for(i = 0; i < PQntuples(res); i++)
+	if (scope->version >= 90100)
 	{
-		fprintf(out, "%-16s  %19s  %13s  %17s  %18s  %17s\n",
-			PQgetvalue(res, i, 0),
-			PQgetvalue(res, i, 1),
-			PQgetvalue(res, i, 2),
-			PQgetvalue(res, i, 3),
-			PQgetvalue(res, i, 4),
-			PQgetvalue(res, i, 5));
+		res = pgut_execute(conn, SQL_SELECT_RECOVERY_CONFLICTS, lengthof(params), params);
+		for(i = 0; i < PQntuples(res); i++)
+		{
+			fprintf(out, "%-16s  %19s  %13s  %17s  %18s  %17s\n",
+				PQgetvalue(res, i, 0),
+				PQgetvalue(res, i, 1),
+				PQgetvalue(res, i, 2),
+				PQgetvalue(res, i, 3),
+				PQgetvalue(res, i, 4),
+				PQgetvalue(res, i, 5));
+		}
+		PQclear(res);
 	}
 	fprintf(out, "\n");
-	PQclear(res);
 }
 
 /*
@@ -348,6 +397,40 @@ report_instance_activity(PGconn *conn, ReportScope *scope, FILE *out)
 	fprintf(out, "/* Instance Activity */\n");
 	fprintf(out, "----------------------------------------\n\n");
 
+	fprintf(out, "/** WAL Statistics **/\n");
+	fprintf(out, "-----------------------------------\n");
+
+	res = pgut_execute(conn, SQL_SELECT_WALSTATS, lengthof(params), params);
+	if (PQntuples(res) == 0)
+		return;
+	if (PQgetisnull(res, 0, 0))
+		fprintf(out, "WAL Write Total  : (N/A)\n");
+	else
+		fprintf(out, "WAL Write Total  : %s MiB\n", PQgetvalue(res, 0, 0));
+	if (PQgetisnull(res, 0, 1))
+		fprintf(out, "WAL Write Speed  : (N/A)\n\n");
+	else
+		fprintf(out, "WAL Write Speed  : %s MiB/s\n\n", PQgetvalue(res, 0, 1));
+	PQclear(res);
+
+	fprintf(out, "-----------------------------------\n");
+	fprintf(out, "%-16s  %-17s  %-24s  %14s  %14s\n",
+		"DateTime", "Location", "Segment File", "Write Size", "Write Size/s");
+	fprintf(out, "------------------------------------------------------------------------------------------------\n");
+
+	res = pgut_execute(conn, SQL_SELECT_WALSTATS_TENDENCY, lengthof(params), params);
+	for(i = 0; i < PQntuples(res); i++)
+	{
+		fprintf(out, "%-16s  %-17s  %-24s  %10s MiB  %10s MiB\n",
+			PQgetvalue(res, i, 0),
+			PQgetvalue(res, i, 1),
+			PQgetvalue(res, i, 2),
+			PQgetvalue(res, i, 3),
+			PQgetvalue(res, i, 4));
+	}
+	fprintf(out, "\n");
+	PQclear(res);
+
 	fprintf(out, "/** Instance Processes Ratio **/\n");
 	fprintf(out, "-----------------------------------\n");
 
@@ -369,7 +452,7 @@ report_instance_activity(PGconn *conn, ReportScope *scope, FILE *out)
 	res = pgut_execute(conn, SQL_SELECT_INSTANCE_PROC_TENDENCY, lengthof(params), params);
 	for(i = 0; i < PQntuples(res); i++)
 	{
-		fprintf(out, "%-16s  %12s  %12s  %12s  %12s\n",
+		fprintf(out, "%-16s  %10s %%  %10s %%  %10s %%  %10s %%\n",
 			PQgetvalue(res, i, 0),
 			PQgetvalue(res, i, 1),
 			PQgetvalue(res, i, 2),
@@ -424,17 +507,35 @@ report_resource_usage(PGconn *conn, ReportScope *scope, FILE *out)
 	fprintf(out, "\n");
 	PQclear(res);
 
+	fprintf(out, "/** Load Average **/\n");
+	fprintf(out, "-----------------------------------\n");
+	fprintf(out, "%-16s  %7s  %7s  %7s\n",
+		"DateTime", "1 Min", "5 Min", "15 Min");
+	fprintf(out, "----------------------------------------------\n");
+
+	res = pgut_execute(conn, SQL_SELECT_LOADAVG_TENDENCY, lengthof(params), params);
+	for(i = 0; i < PQntuples(res); i++)
+	{
+		fprintf(out, "%-16s  %7s  %7s  %7s\n",
+			PQgetvalue(res, i, 0),
+			PQgetvalue(res, i, 1),
+			PQgetvalue(res, i, 2),
+			PQgetvalue(res, i, 3));
+	}
+	fprintf(out, "\n");
+	PQclear(res);
+
 	fprintf(out, "/** IO Usage **/\n");
 	fprintf(out, "-----------------------------------\n");
-	fprintf(out, "%-12s  %-24s  %11s  %11s  %17s  %17s  %16s  %15s\n",
+	fprintf(out, "%-12s  %-24s  %12s  %12s  %17s  %17s  %16s  %15s\n",
 		"Device", "Including TabelSpaces", "Total Read", "Total Write",
 		"Total Read Time", "Total Write Time", "Current IO Queue", "Total IO Time");
-	fprintf(out, "--------------------------------------------------------------------------------------------------------------------------------------------\n");
+	fprintf(out, "----------------------------------------------------------------------------------------------------------------------------------------------\n");
 
 	res = pgut_execute(conn, SQL_SELECT_IO_USAGE, lengthof(params), params);
 	for(i = 0; i < PQntuples(res); i++)
 	{
-		fprintf(out, "%-12s  %-24s  %8s MB  %8s MB  %14s ms  %14s ms  %16s  %12s ms\n",
+		fprintf(out, "%-12s  %-24s  %8s MiB  %8s MiB  %14s ms  %14s ms  %16s  %12s ms\n",
 			PQgetvalue(res, i, 0),
 			PQgetvalue(res, i, 1),
 			PQgetvalue(res, i, 2),
@@ -448,14 +549,34 @@ report_resource_usage(PGconn *conn, ReportScope *scope, FILE *out)
 	PQclear(res);
 
 	fprintf(out, "-----------------------------------\n");
-	fprintf(out, "%-16s  %-12s  %15s  %15s  %15s  %15s\n",
+	fprintf(out, "%-16s  %-12s  %14s  %14s  %15s  %15s\n",
 		"DateTime", "Device", "Read Size/s", "Write Size/s", "Read Time/s", "Write Time/s");
-	fprintf(out, "-----------------------------------------------------------------------------------------------------\n");
+	fprintf(out, "---------------------------------------------------------------------------------------------------\n");
 
 	res = pgut_execute(conn, SQL_SELECT_IO_USAGE_TENDENCY, lengthof(params), params);
 	for(i = 0; i < PQntuples(res); i++)
 	{
-		fprintf(out, "%-16s  %-12s  %12s KB  %12s KB  %12s ms  %12s ms\n",
+		fprintf(out, "%-16s  %-12s  %10s KiB  %10s KiB  %12s ms  %12s ms\n",
+			PQgetvalue(res, i, 0),
+			PQgetvalue(res, i, 1),
+			PQgetvalue(res, i, 2),
+			PQgetvalue(res, i, 3),
+			PQgetvalue(res, i, 4),
+			PQgetvalue(res, i, 5));
+	}
+	fprintf(out, "\n");
+	PQclear(res);
+
+	fprintf(out, "/** Memory Usage **/\n");
+	fprintf(out, "-----------------------------------\n");
+	fprintf(out, "%-16s  %12s  %12s  %12s  %12s  %12s\n",
+		"DateTime", "Memfree", "Buffers", "Cached", "Swap", "Dirty");
+	fprintf(out, "-----------------------------------------------------------------------------------------\n");
+
+	res = pgut_execute(conn, SQL_SELECT_MEMORY_TENDENCY, lengthof(params), params);
+	for(i = 0; i < PQntuples(res); i++)
+	{
+		fprintf(out, "%-16s  %8s MiB  %8s MiB  %8s MiB  %8s MiB  %8s MiB\n",
 			PQgetvalue(res, i, 0),
 			PQgetvalue(res, i, 1),
 			PQgetvalue(res, i, 2),
@@ -483,14 +604,14 @@ report_disk_usage(PGconn *conn, ReportScope *scope, FILE *out)
 
 	fprintf(out, "/** Disk Usage per Tablespace **/\n");
 	fprintf(out, "-----------------------------------\n");
-	fprintf(out, "%-16s  %-32s  %-12s  %11s  %11s  %10s\n",
+	fprintf(out, "%-16s  %-32s  %-12s  %12s  %12s  %10s\n",
 		"Tablespace", "Location", "Device", "Used", "Avail", "Remain");
-	fprintf(out, "---------------------------------------------------------------------------------------------------------\n");
+	fprintf(out, "-----------------------------------------------------------------------------------------------------------\n");
 
 	res = pgut_execute(conn, SQL_SELECT_DISK_USAGE_TABLESPACE, lengthof(params), params);
 	for(i = 0; i < PQntuples(res); i++)
 	{
-		fprintf(out, "%-16s  %-32s  %-12s  %8s MB  %8s MB  %8s %%\n",
+		fprintf(out, "%-16s  %-32s  %-12s  %8s MiB  %8s MiB  %8s %%\n",
 			PQgetvalue(res, i, 0),
 			PQgetvalue(res, i, 1),
 			PQgetvalue(res, i, 2),
@@ -503,14 +624,14 @@ report_disk_usage(PGconn *conn, ReportScope *scope, FILE *out)
 
 	fprintf(out, "/** Disk Usage per Table **/\n");
 	fprintf(out, "-----------------------------------\n");
-	fprintf(out, "%-16s  %-16s  %-16s  %11s  %12s  %12s  %12s\n",
+	fprintf(out, "%-16s  %-16s  %-16s  %12s  %12s  %12s  %12s\n",
 		"Database", "Schema", "Table", "Size", "Table Reads", "Index Reads", "Toast Reads");
-	fprintf(out, "--------------------------------------------------------------------------------------------------------------\n");
+	fprintf(out, "---------------------------------------------------------------------------------------------------------------\n");
 
 	res = pgut_execute(conn, SQL_SELECT_DISK_USAGE_TABLE, lengthof(params), params);
 	for(i = 0; i < PQntuples(res); i++)
 	{
-		fprintf(out, "%-16s  %-16s  %-16s  %8s MB  %12s  %12s  %12s\n",
+		fprintf(out, "%-16s  %-16s  %-16s  %8s MiB  %12s  %12s  %12s\n",
 			PQgetvalue(res, i, 0),
 			PQgetvalue(res, i, 1),
 			PQgetvalue(res, i, 2),
@@ -693,28 +814,51 @@ report_autovacuum_activity(PGconn *conn, ReportScope *scope, FILE *out)
 
 	fprintf(out, "----------------------------------------\n");
 	fprintf(out, "/* Autovacuum Activity */\n");
-	fprintf(out, "----------------------------------------\n");
-	fprintf(out, "%-16s  %-16s  %-16s  %8s  %16s  %17s  %16s  %15s  %15s\n",
-		"Database", "Schema", "Table", "Count", "Index Scans(Avg)",
-		"Removed Rows(Avg)", "Remain Rows(Avg)", "Duration(Avg)", "Duration(Max)");
-	fprintf(out, "----------------------------------------------------------------------------------------------------------------------------------------------------------\n");
+	fprintf(out, "----------------------------------------\n\n");
+
+	fprintf(out, "/** Basic Statistics (Average) **/\n");
+	fprintf(out, "-----------------------------------\n");
+	fprintf(out, "%-32s  %8s  %12s  %12s  %12s  %12s  %13s\n",
+		"Table", "Count", "Index Scans", "Removed Rows", "Remain Rows", "Duration", "Duration(Max)");
+	fprintf(out, "--------------------------------------------------------------------------------------------------------------------\n");
 
 	res = pgut_execute(conn, SQL_SELECT_AUTOVACUUM_ACTIVITY, lengthof(params), params);
 	for(i = 0; i < PQntuples(res); i++)
 	{
-		fprintf(out, "%-16s  %-16s  %-16s  %8s  %16s  %17s  %16s  %13s s  %13s s\n",
+		fprintf(out, "%-32s  %8s  %12s  %12s  %12s  %10s s  %11s s\n",
 			PQgetvalue(res, i, 0),
 			PQgetvalue(res, i, 1),
 			PQgetvalue(res, i, 2),
 			PQgetvalue(res, i, 3),
 			PQgetvalue(res, i, 4),
 			PQgetvalue(res, i, 5),
-			PQgetvalue(res, i, 6),
-			PQgetvalue(res, i, 7),
-			PQgetvalue(res, i, 8));
+			PQgetvalue(res, i, 6));
 	}
 	fprintf(out, "\n");
 	PQclear(res);
+
+	if (scope->version >= 90200)
+	{
+		fprintf(out, "/** I/O Statistics (Average) **/\n");
+		fprintf(out, "-----------------------------------\n");
+		fprintf(out, "%-32s  %10s  %10s  %10s  %13s  %13s\n",
+			"Table", "Page Hit", "Page Miss", "Page Dirty", "Read Rate", "Write Rate");
+		fprintf(out, "-----------------------------------------------------------------------------------------------------\n");
+
+		res = pgut_execute(conn, SQL_SELECT_AUTOVACUUM_ACTIVITY2, lengthof(params), params);
+		for(i = 0; i < PQntuples(res); i++)
+		{
+			fprintf(out, "%-32s  %10s  %10s  %10s  %7s MiB/s  %7s MiB/s\n",
+				PQgetvalue(res, i, 0),
+				PQgetvalue(res, i, 1),
+				PQgetvalue(res, i, 2),
+				PQgetvalue(res, i, 3),
+				PQgetvalue(res, i, 4),
+				PQgetvalue(res, i, 5));
+		}
+		fprintf(out, "\n");
+		PQclear(res);
+	}
 }
 
 /*
@@ -756,44 +900,54 @@ report_query_activity(PGconn *conn, ReportScope *scope, FILE *out)
 
 	fprintf(out, "/** Statements **/\n");
 	fprintf(out, "-----------------------------------\n");
-	fprintf(out, "%-16s  %-16s  %8s  %13s  %12s  %-s\n",
-		"User", "Database", "Calls", "Total Time", "Time/Call", "Query");
-	fprintf(out, "--------------------------------------------------------------------------------------------\n");
+	fprintf(out, "%-16s  %-16s  %8s  %14s  %13s  %15s  %16s  %-s\n",
+		"User", "Database", "Calls", "Total Time", "Time/Call",
+		"Block Read Time", "Block Write Time", "Query");
+	fprintf(out, "------------------------------------------------------------------------------------------------------------------------\n");
 
 	res = pgut_execute(conn, SQL_SELECT_QUERY_ACTIVITY_STATEMENTS, lengthof(params), params);
 	for(i = 0; i < PQntuples(res); i++)
 	{
-		fprintf(out, "%-16s  %-16s  %8s  %10s ms  %9s ms  %-s\n",
-			PQgetvalue(res, i, 0),
-			PQgetvalue(res, i, 1),
-			PQgetvalue(res, i, 3),
-			PQgetvalue(res, i, 4),
-			PQgetvalue(res, i, 5),
-			PQgetvalue(res, i, 2));
+		fprintf(out, "%-16s  ", PQgetvalue(res, i, 0));
+		fprintf(out, "%-16s  ", PQgetvalue(res, i, 1));
+		fprintf(out, "%8s  ", PQgetvalue(res, i, 3));
+		fprintf(out, "%10s sec  ", PQgetvalue(res, i, 4));
+		fprintf(out, "%9s sec  ", PQgetvalue(res, i, 5));
+		if (scope->version >= 90200)
+		{
+			fprintf(out, "%12s ms  ", PQgetvalue(res, i, 6));
+			fprintf(out, "%13s ms  ", PQgetvalue(res, i, 7));
+		}
+		else
+		{
+			fprintf(out, "%15s  ", "(N/A)");
+			fprintf(out, "%16s  ", "(N/A)");
+		}
+		fprintf(out, "%-s\n", PQgetvalue(res, i, 2));
 	}
 	fprintf(out, "\n");
 	PQclear(res);
 }
 
 /*
- * generate a report that corresponds to 'LockActivity'
+ * generate a report that corresponds to 'LockConflicts'
  */
 static void
-report_lock_activity(PGconn *conn, ReportScope *scope, FILE *out)
+report_lock_conflicts(PGconn *conn, ReportScope *scope, FILE *out)
 {
 	PGresult	*res;
 	const char	*params[] = { scope->beginid, scope->endid };
 	int			 i;
 
 	fprintf(out, "----------------------------------------\n");
-	fprintf(out, "/* Lock Activity */\n");
+	fprintf(out, "/* Lock Conflicts */\n");
 	fprintf(out, "----------------------------------------\n");
 	fprintf(out, "%-16s  %-16s  %-16s  %-8s  %11s  %11s  %-16s\n%-s\n%-s\n",
 		"Database", "Schema", "Relation", "Duration", "Blockee PID", "Blocker PID", "Blocker GID",
 		"Blockee Query", "Blocker Query");
 	fprintf(out, "--------------------------------------------------------------------------------------------------------\n");
 
-	res = pgut_execute(conn, SQL_SELECT_LOCK_ACTIVITY, lengthof(params), params);
+	res = pgut_execute(conn, SQL_SELECT_LOCK_CONFLICTS, lengthof(params), params);
 	for(i = 0; i < PQntuples(res); i++)
 	{
 		fprintf(out, "%-16s  %-16s  %-16s  %-8s  %11s  %11s  %-16s\n%-s\n%-s\n",
@@ -841,8 +995,9 @@ report_replication_activity(PGconn *conn, ReportScope *scope, FILE *out)
 		fprintf(out, "Flush WAL Location    : %s\n", PQgetvalue(res, i, 10));
 		fprintf(out, "Replay WAL Location   : %s\n", PQgetvalue(res, i, 11));
 		fprintf(out, "Sync Priority         : %s\n", PQgetvalue(res, i, 12));
-		fprintf(out, "Sync State            : %s\n\n", PQgetvalue(res, i, 13));
+		fprintf(out, "Sync State            : %s\n", PQgetvalue(res, i, 13));
 	}
+	fprintf(out, "\n");
 	PQclear(res);
 }
 
@@ -891,15 +1046,15 @@ report_schema_information(PGconn *conn, ReportScope *scope, FILE *out)
 
 	fprintf(out, "/** Tables **/\n");
 	fprintf(out, "-----------------------------------\n");
-	fprintf(out, "%-16s  %-16s  %-16s  %8s  %14s  %9s  %12s  %11s  %11s\n",
+	fprintf(out, "%-16s  %-16s  %-16s  %8s  %14s  %10s  %10s  %11s  %11s\n",
 		"Database", "Schema", "Table", "Columns", "Row Width", "Size",
 		"Size Incr", "Table Scans", "Index Scans");
-	fprintf(out, "------------------------------------------------------------------------------------------------------------------------------------\n");
+	fprintf(out, "-----------------------------------------------------------------------------------------------------------------------------------\n");
 
 	res = pgut_execute(conn, SQL_SELECT_SCHEMA_INFORMATION_TABLES, lengthof(params), params);
 	for(i = 0; i < PQntuples(res); i++)
 	{
-		fprintf(out, "%-16s  %-16s  %-16s  %8s  %9s byte  %6s MB  %9s MB  %11s  %11s\n",
+		fprintf(out, "%-16s  %-16s  %-16s  %8s  %9s byte  %6s MiB  %6s MiB  %11s  %11s\n",
 			PQgetvalue(res, i, 0),	
 			PQgetvalue(res, i, 1),
 			PQgetvalue(res, i, 2),
@@ -915,15 +1070,15 @@ report_schema_information(PGconn *conn, ReportScope *scope, FILE *out)
 
 	fprintf(out, "/** Indexes **/\n");
 	fprintf(out, "-----------------------------------\n");
-	fprintf(out, "%-16s  %-16s  %-16s  %-16s  %9s  %12s  %11s  %9s  %10s  %11s  %-s\n",
+	fprintf(out, "%-16s  %-16s  %-16s  %-16s  %10s  %10s  %11s  %9s  %10s  %11s  %-s\n",
 		"Database", "Schema", "Index", "Table", "Size", "Size Incr",
 		"Index Scans", "Rows/Scan", "Disk Reads", "Cache Reads", "Index Key");
-	fprintf(out, "--------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
+	fprintf(out, "-------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
 
 	res = pgut_execute(conn, SQL_SELECT_SCHEMA_INFORMATION_INDEXES, lengthof(params), params);
 	for(i = 0; i < PQntuples(res); i++)
 	{
-		fprintf(out, "%-16s  %-16s  %-16s  %-16s  %6s MB  %9s MB  %11s  %9s  %10s  %11s  %-s\n",
+		fprintf(out, "%-16s  %-16s  %-16s  %-16s  %6s MiB  %6s MiB  %11s  %9s  %10s  %11s  %-s\n",
 			PQgetvalue(res, i, 0),	
 			PQgetvalue(res, i, 1),
 			PQgetvalue(res, i, 2),
@@ -984,7 +1139,7 @@ report_all(PGconn *conn, ReportScope *scope, FILE *out)
 	report_checkpoint_activity(conn, scope, out);
 	report_autovacuum_activity(conn, scope, out);
 	report_query_activity(conn, scope, out);
-	report_lock_activity(conn, scope, out);
+	report_lock_conflicts(conn, scope, out);
 	report_replication_activity(conn, scope, out);
 	report_setting_parameters(conn, scope, out);
 	report_schema_information(conn, scope, out);
@@ -1032,8 +1187,8 @@ parse_reportid(const char *value)
 		return (ReportBuild) report_autovacuum_activity;
 	else if (pg_strncasecmp(REPORTID_QUERY_ACTIVITY, v, len) == 0)
 		return (ReportBuild) report_query_activity;
-	else if (pg_strncasecmp(REPORTID_LOCK_ACTIVITY, v, len) == 0)
-		return (ReportBuild) report_lock_activity;
+	else if (pg_strncasecmp(REPORTID_LOCK_CONFLICTS, v, len) == 0)
+		return (ReportBuild) report_lock_conflicts;
 	else if (pg_strncasecmp(REPORTID_REPLICATION_ACTIVITY, v, len) == 0)
 		return (ReportBuild) report_replication_activity;
 	else if (pg_strncasecmp(REPORTID_SETTING_PARAMETERS, v, len) == 0)
@@ -1085,8 +1240,9 @@ select_scope_by_snapid(PGconn *conn, const char *beginid, const char *endid)
 		scope->instid = pgut_strdup(PQgetvalue(res, i, 0));		/* instance ID */
 		scope->host = pgut_strdup(PQgetvalue(res, i, 1));		/* host */
 		scope->port = pgut_strdup(PQgetvalue(res, i, 2));		/* port */
-		scope->beginid = pgut_strdup(PQgetvalue(res, i, 3));	/* begin point of report */
-		scope->endid = pgut_strdup(PQgetvalue(res, i, 4));		/* end point of report */
+		scope->version = parse_version(PQgetvalue(res, i, 3));	/* PostgreSQL version */
+		scope->beginid = pgut_strdup(PQgetvalue(res, i, 4));	/* begin point of report */
+		scope->endid = pgut_strdup(PQgetvalue(res, i, 5));		/* end point of report */
 		scope_list = lappend(scope_list, scope);
 	}
 	PQclear(res);
@@ -1129,8 +1285,9 @@ select_scope_by_timestamp(PGconn *conn, time_t begindate, time_t enddate)
 		scope->instid = pgut_strdup(PQgetvalue(res, i, 0));		/* instance ID */
 		scope->host = pgut_strdup(PQgetvalue(res, i, 1));		/* host */
 		scope->port = pgut_strdup(PQgetvalue(res, i, 2));		/* port */
-		scope->beginid = pgut_strdup(PQgetvalue(res, i, 3));	/* begin point of report */
-		scope->endid = pgut_strdup(PQgetvalue(res, i, 4));		/* end point of report */
+		scope->version = parse_version(PQgetvalue(res, i, 3));	/* PostgreSQL version */
+		scope->beginid = pgut_strdup(PQgetvalue(res, i, 4));	/* begin point of report */
+		scope->endid = pgut_strdup(PQgetvalue(res, i, 5));		/* end point of report */
 		scope_list = lappend(scope_list, scope);
 	}
 	PQclear(res);
@@ -1163,3 +1320,20 @@ get_server_version(PGconn *conn)
 
 	return server_version_num;
 }
+
+static int
+parse_version(const char *versionString)
+{
+	int	cnt;
+	int	vmaj, vmin, vrev;
+
+	cnt = sscanf(versionString, "%d.%d.%d", &vmaj, &vmin, &vrev);
+
+	if (cnt < 2)
+		return -1;
+
+	if (cnt == 2)
+		vrev = 0;
+
+	return (100 * vmaj + vmin) * 100 + vrev;
+}
